/*
 * double2s32.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * Copyright (C) 2000-2002 Michel Lespinasse <walken@zoy.org>
 * Copyright (C) 1999-2000 Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <a52dec/config.h>

#include <stdio.h>
#include <inttypes.h>

#include <a52dec/a52.h>
#include <a52dec/audio_out.h>

#include "audio_out_internal.h"

static inline int32_t convert (int64_t i)
{
    if (i > 0x414800007fffffffll)
	return 2147483647ll;
    else if (i < 0x4147ffff80000000ll)
	return -2147483648ll;
    else
	return i - 0x4148000000000000ll;
}

void double2s32_2 (double * _f, int32_t * s32)
{
    int i;
    int64_t * f = (int64_t *) _f;

    for (i = 0; i < 256; i++) {
	s32[2*i] = convert (f[i]);
	s32[2*i+1] = convert (f[i+256]);
    }
}

void double2s32_4 (double * _f, int32_t * s32)
{
    int i;
    int64_t * f = (int64_t *) _f;

    for (i = 0; i < 256; i++) {
	s32[4*i] = convert (f[i]);
	s32[4*i+1] = convert (f[i+256]);
	s32[4*i+2] = convert (f[i+512]);
	s32[4*i+3] = convert (f[i+768]);
    }
}

void double2s32_5 (double * _f, int32_t * s32)
{
    int i;
    int64_t * f = (int64_t *) _f;

    for (i = 0; i < 256; i++) {
	s32[5*i] = convert (f[i]);
	s32[5*i+1] = convert (f[i+256]);
	s32[5*i+2] = convert (f[i+512]);
	s32[5*i+3] = convert (f[i+768]);
	s32[5*i+4] = convert (f[i+1024]);
    }
}

int channels_multi (int flags)
{
    if (flags & A52_LFE)
	return 6;
    else if (flags & 1)	/* center channel */
	return 5;
    else if ((flags & A52_CHANNEL_MASK) == A52_2F2R)
	return 4;
    else
	return 2;
}

void double2s32_multi (double * _f, int32_t * s32, int flags)
{
    int i;
    int64_t * f = (int64_t *) _f;

    switch (flags) {
    case A52_MONO:
	for (i = 0; i < 256; i++) {
	    s32[5*i] = s32[5*i+1] = s32[5*i+2] = s32[5*i+3] = 0;
	    s32[5*i+4] = convert (f[i]);
	}
	break;
    case A52_CHANNEL:
    case A52_STEREO:
    case A52_DOLBY:
	double2s32_2 (_f, s32);
	break;
    case A52_3F:
	for (i = 0; i < 256; i++) {
	    s32[5*i] = convert (f[i]);
	    s32[5*i+1] = convert (f[i+512]);
	    s32[5*i+2] = s32[5*i+3] = 0;
	    s32[5*i+4] = convert (f[i+256]);
	}
	break;
    case A52_2F2R:
	double2s32_4 (_f, s32);
	break;
    case A52_3F2R:
	double2s32_5 (_f, s32);
	break;
    case A52_MONO | A52_LFE:
	for (i = 0; i < 256; i++) {
	    s32[6*i] = s32[6*i+1] = s32[6*i+2] = s32[6*i+3] = 0;
	    s32[6*i+4] = convert (f[i+256]);
	    s32[6*i+5] = convert (f[i]);
	}
	break;
    case A52_CHANNEL | A52_LFE:
    case A52_STEREO | A52_LFE:
    case A52_DOLBY | A52_LFE:
	for (i = 0; i < 256; i++) {
	    s32[6*i] = convert (f[i+256]);
	    s32[6*i+1] = convert (f[i+512]);
	    s32[6*i+2] = s32[6*i+3] = s32[6*i+4] = 0;
	    s32[6*i+5] = convert (f[i]);
	}
	break;
    case A52_3F | A52_LFE:
	for (i = 0; i < 256; i++) {
	    s32[6*i] = convert (f[i+256]);
	    s32[6*i+1] = convert (f[i+768]);
	    s32[6*i+2] = s32[6*i+3] = 0;
	    s32[6*i+4] = convert (f[i+512]);
	    s32[6*i+5] = convert (f[i]);
	}
	break;
    case A52_2F2R | A52_LFE:
	for (i = 0; i < 256; i++) {
	    s32[6*i] = convert (f[i+256]);
	    s32[6*i+1] = convert (f[i+512]);
	    s32[6*i+2] = convert (f[i+768]);
	    s32[6*i+3] = convert (f[i+1024]);
	    s32[6*i+4] = 0;
	    s32[6*i+5] = convert (f[i]);
	}
	break;
    case A52_3F2R | A52_LFE:
	for (i = 0; i < 256; i++) {
	    s32[6*i] = convert (f[i+256]);
	    s32[6*i+1] = convert (f[i+768]);
	    s32[6*i+2] = convert (f[i+1024]);
	    s32[6*i+3] = convert (f[i+1280]);
	    s32[6*i+4] = convert (f[i+512]);
	    s32[6*i+5] = convert (f[i]);
	}
	break;
    }
}

#ifdef __XXX__
void s32_swap (int32_t * s32, int channels)
{
    int i;
    uint32_t * u16 = (uint32_t *) s32;

    for (i = 0; i < 256 * channels; i++)
	u16[i] = (u16[i] >> 8) | (u16[i] << 8);
}
#endif
