/*
 * in_ffsox.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"
#include "vout.h"
#define __DYNAMIC__
#include "ffmpeg_stub.h"
#include "libsox_stub.h"

//#define __SYNC__

typedef struct stream {
  int id;
  AVStream *stream;
  AVCodecContext *dec;
  AVCodec *codec;
  double time_base;
} stream_t;

typedef struct audio {
  stream_t s;
  sox_effects_chain_t *chain;
  AVPacket p1;
  AVPacket p2;
  uint8_t *rp;
  uint8_t *mp;
  int start;
  int eof;
  void (*settime)(int);
  int (*read_frame)(track_t *, AVPacket *);
  int done;
} audio_t;

typedef struct video {
  stream_t s;
  volatile int run;
  volatile int pause;
  HANDLE hEventWrite;
  packet_queue_t qa;
  packet_queue_t qv;

  int width;
  int height;
  int in_pix_fmt;
  int out_pix_fmt;

  vout_t *vout;
  int (*gettime)();
  double step;
  double pts;
  timer_t timer;
  AVFrame *frame;
  struct SwsContext *img_convert_ctx;

  AVPicture out_pict;
  AVPicture in_pict;

#ifdef __SYNC__
  struct {
    CRITICAL_SECTION cs;
    HANDLE hEvent;
    int count;
    int a;
    int v;
  } sync;
#endif // __SYNC__
} video_t;

struct track {
  AVFormatContext *ic;
  int64_t start_time;
  audio_t a;
  video_t v;
  samplebuf_t *buf;

  HANDLE hThreadSoX;
};

typedef int (*decode_t)(AVCodecContext *, void *, int *, AVPacket *);

static track_t track_play;
static track_t track_info;
DECLARE_ALIGNED(16,uint8_t,ff_samples[(AVCODEC_MAX_AUDIO_FRAME_SIZE*3)/2]);

static void get_tags_md(AVMetadata *metadata, samplebuf_t *buf)
{
  AVMetadataTag *tag=NULL;

#define MAKE_INFO(id) \
  if (NULL!=(tag=av_metadata_get(metadata,#id,NULL,\
      AV_METADATA_IGNORE_SUFFIX))) \
    _snprintf (buf->tag.id, sizeof buf->tag.id, "%s", tag->value);
#define MAKE_ALTERNATE_INFO(id1, id2) \
  if (0==buf->tag.id1[0]&&NULL!=(tag=av_metadata_get(metadata,#id2,NULL,\
      AV_METADATA_IGNORE_SUFFIX))) \
    _snprintf (buf->tag.id1, sizeof buf->tag.id1, "%s", tag->value);
#include "info.h"
#undef MAKE_ALTERNATE_INFO
#undef MAKE_INFO
}

static void in_ffsox_fill_format(track_t *track)
{
  samplebuf_t *buf=track->buf;
  video_t *v=&track->v;
  audio_t *a=&track->a;

  if (NULL!=v->s.codec) {
    _snprintf(buf->formatinformation,
        sizeof buf->formatinformation,
        "Length: %d seconds"
        "\nDecoder: FFmpeg"
        "\n\nAudio Codec: %s"
        "\nChannels: %d"
        "\nBits per sample (in): %d"
        "\nSample Rate (in): %.f Hz"
        "\nBits per sample (out): %d"
        "\nSample Rate (out): %.f Hz"
        //"\nFile Size: %d bytes"
        //"\nAverage bitrate: %d kbps"
        //"\nCompression Ratio: %d%%"
        "\n\nVideo Codec: %s"
        ,
        buf->length/1000,
        a->s.codec?a->s.codec->name:"undefined",
        a->s.dec->channels,
        0<a->s.dec->bits_per_raw_sample
                ?a->s.dec->bits_per_raw_sample
                :av_get_bits_per_sample_format(a->s.dec->sample_fmt),
        (double)a->s.dec->sample_rate,
        track->buf->bps,
        (double)track->buf->rate,
        v->s.codec?v->s.codec->name:"undefined");
  }
  else {
    _snprintf(buf->formatinformation,
        sizeof buf->formatinformation,
        "Length: %d seconds"
        "\nDecoder: FFmpeg"
        "\n\nAudio Codec: %s"
        "\nChannels: %d"
        "\nBits per sample (in): %d"
        "\nSample Rate (in): %.f Hz"
        "\nBits per sample (out): %d"
        "\nSample Rate (out): %.f Hz"
        //"\nFile Size: %d bytes"
        //"\nAverage bitrate: %d kbps"
        //"\nCompression Ratio: %d%%"
        ,
        buf->length/1000,
        a->s.codec?a->s.codec->name:"undefinded",
        a->s.dec->channels,
        0<a->s.dec->bits_per_raw_sample
                ?a->s.dec->bits_per_raw_sample
                :av_get_bits_per_sample_format(a->s.dec->sample_fmt),
        (double)a->s.dec->sample_rate,
        track->buf->bps,
        (double)track->buf->rate);
  }
}

static void get_tags_fn(samplebuf_t *buf)
{
  char tmp[256];
  char *mp=tmp+sizeof(tmp)-1;
  const char *fn=buf->fn+strlen(buf->fn);
  int offs;
  FILE *csv=NULL;
  int ch;

  while (buf->fn<fn&&'\\'!=fn[-1]&&'/'!=fn[-1])
    --fn;

  offs=fn - buf->fn;
  memcpy(buf->csv, buf->fn, offs);
  strcpy(buf->csv+offs, "folder.csv");

  if (NULL==(csv=fopen(buf->csv, "rb")))
    return;

  while ('\n'!=(ch=fgetc(csv))) {
    if (EOF==ch)
      goto cleanup;
  }

  for (;;) {
    char *wp=tmp;

    // File
    while ('\t'!=(ch=fgetc(csv))) {
      if (EOF==ch||wp==mp)
        goto cleanup;
      else
        *wp++=ch;
    }

    *wp=0;

    if (0==strcmp(fn, tmp)) {
#define MAKE_INFO(id) \
      wp=buf->tag.id; \
      mp=buf->tag.id+sizeof(buf->tag.id)-1; \
 \
      for (;;) { \
        ch=fgetc(csv); \
 \
        if (wp+1==mp) \
          goto cleanup; \
        else if ('\t'==ch||'\n'==ch||EOF==ch) \
          break; \
        else \
          *wp++=ch; \
      } \
 \
      *wp=0;
#define MAKE_ALTERNATE_INFO(id1, id2)
#include "info.h"
#undef MAKE_ALTERNATE_INFO
#undef MAKE_INFO
      break;
    }
    else {
      while ('\n'!=(ch=fgetc(csv))) {
        if (EOF==ch)
          goto cleanup;
      }
    }
  }
cleanup:
  if (NULL!=csv)
    fclose(csv);
}

#ifdef __SYNC__
static void in_ffsox_sync(track_t *track, int *guard)
{
  video_t *v=&track->v;

  if (NULL!=v->s.codec && 0==*guard) {
    EnterCriticalSection(&v->sync.cs);

    if (++v->sync.count<2) {
      LeaveCriticalSection(&v->sync.cs);
      WaitForSingleObject(&v->sync.hEvent, INFINITE);
      EnterCriticalSection(&v->sync.cs);
    }
    else
      SetEvent(&v->sync.hEvent);

    LeaveCriticalSection(&v->sync.cs);
    *guard=1;
  }
}
#endif // __SYNC__

static void in_ffsox_decode_video(track_t *track, AVPacket *p1)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;
  AVFrame *frame=v->frame;
  int got_picture=0;
  int64_t pts;

  v->s.dec->reordered_opaque=p1->pts;
  avcodec_decode_video2(v->s.dec, frame, &got_picture, p1);

  if (got_picture) {
    if (NULL==v->img_convert_ctx) {
      if (! v->frame->interlaced_frame) {
        av_picture_copy(&v->out_pict, (AVPicture *)v->frame,
            v->in_pix_fmt, v->width, v->height);
      }
      else {
        avpicture_deinterlace(&v->out_pict, (AVPicture *)v->frame,
            v->in_pix_fmt, v->width, v->height);
      }
    }
    else if (! v->frame->interlaced_frame) {
      sws_scale(v->img_convert_ctx, v->frame->data, v->frame->linesize,
          0, v->height, v->out_pict.data, v->out_pict.linesize);
    }
    else {
      avpicture_deinterlace(&v->in_pict, (AVPicture *)v->frame,
          v->in_pix_fmt, v->width, v->height);
      sws_scale(v->img_convert_ctx, v->in_pict.data, v->in_pict.linesize,
          0, v->height, v->out_pict.data, v->out_pict.linesize);
    }

    if (p1->dts!=AV_NOPTS_VALUE) {
      pts=p1->dts;
#ifdef __DEBUG__
//#define __XDEBUG__
#endif
#ifdef __XDEBUG__
      fprintf(stderr, "!!! 1: pts: %.0f, %.0f\n", (double)pts,
          (double)stream->start_time);
#endif
    }
    else if (v->frame->reordered_opaque!=AV_NOPTS_VALUE) {
      pts=v->frame->reordered_opaque;
#ifdef __XDEBUG__
      fprintf(stderr, "!!! 2: pts: %.0f\n", (double)pts);
#endif
    }
    else {
      pts=v->pts;
      v->pts+=v->step;
#ifdef __XDEBUG__
      fprintf(stderr, "!!! 3: pts: %.0f\n", (double)pts);
#endif
    }

#ifdef __SYNC__
    in_ffsox_sync(track, &v->sync.v);
#endif // __SYNC__
    double delay=v->s.time_base*(pts-track->start_time) - v->gettime();

	if (! ini.plugin.drop_frames.on
	    || (ini.plugin.drop_frames.before<=delay
	    && delay<=ini.plugin.drop_frames.after)) {

      if (!a->done&&0.0<delay)
        timer_sleep(&v->timer, delay);

      vout_draw(v->vout,
          v->out_pict.data[0], v->out_pict.linesize[0],
          v->out_pict.data[2], v->out_pict.linesize[2],
          v->out_pict.data[1], v->out_pict.linesize[1]);
	}
  }
}

static void in_ffsox_decode_audio(track_t *track, decode_t decode)
{
  audio_t *a=&track->a;
  AVPacket *p2=&track->a.p2;
  int frame_size = sizeof(ff_samples);
  int n=decode(a->s.dec, (int16_t *)ff_samples, &frame_size, p2);

  if (n<0) {
    p2->size=0;
    p2->data=NULL;
  }
  else {
    p2->size-=n;
    p2->data+=n;
/*
    fwrite(ff_samples, frame_size, 1, stdout);
*/
    a->rp=ff_samples;
    a->mp=ff_samples+frame_size;
  }
}

static int in_ffsox_convert_audio(track_t *track, sox_sample_t *samples,
    size_t len)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;
  int bps=av_get_bits_per_sample_format(a->s.dec->sample_fmt);
  int bytes=bps/8;
  int avail=(a->mp-a->rp)/bytes;
  sox_sample_t *wp=samples;
  sox_sample_t *mp=samples+(len<avail?len:avail);

  while (wp<mp) {
    *wp++=16==bps?SOX_SIGNED_16BIT_TO_SAMPLE(*(int16_t *)a->rp,effp->clips):
        SOX_SIGNED_32BIT_TO_SAMPLE(*(int32_t *)a->rp,effp->clips);
    a->rp+=bytes;
  }

#ifdef __SYNC__
  in_ffsox_sync(track, &v->sync.a);
#endif // __SYNC__

  return wp-samples;
}

static int in_ffsox_read_frame_a(track_t *track, AVPacket *p1)
{
   return av_read_frame(track->ic, p1);
}

static int in_ffsox_read_frame_v(track_t *track, AVPacket *p1)
{
   return packet_queue_read(&track->v.qa, p1)?0:-1;
}

static int in_ffsox_read(sox_effect_t * effp, sox_sample_t *samples, size_t len)
{
  track_t *track=*(track_t **)effp->priv;
  audio_t *a=&track->a;
  AVPacket *p1=&a->p1;
  AVPacket *p2=&a->p2;

  for (;;) {
    if (a->rp<a->mp)
      return in_ffsox_convert_audio(track, samples, len);
    else if (p2->size<0||a->eof)
      return 0;
    else if (0==p2->size) {
      if (NULL!=p1->data) {
        av_free_packet(p1);
        memset(p1, 0, sizeof *p1);
      }

      if (a->read_frame(track, p1)<0) {
        a->eof=1;

        if (NULL==a->s.codec->decode)
          return 0;
		else if (get_mpc7_decoder()==a->s.codec)
		  return 0;
        else
          in_ffsox_decode_audio(track, a->s.codec->decode);
      }
      else if(p1->stream_index==a->s.id) {
        if (!a->start) {
		  if (AV_NOPTS_VALUE!=p1->pts)
		    a->settime(a->s.time_base*(p1->pts-track->start_time));

		  a->start=1;
        }

        p2->data=p1->data;
        p2->size=p1->size;
      }
      else {
        av_free_packet(p1);
        memset(p1, 0, sizeof *p1);
      }
    }
    else
      in_ffsox_decode_audio(track, (decode_t)avcodec_decode_audio3);
  }
}

/* The function that will be called to input samples into the effects chain.
 * In this example, we get samples to process from a SoX-openned audio file.
 * In a different application, they might be generated or come from a different
 * part of the application. */
static int input_drain(
    sox_effect_t * effp, sox_sample_t * obuf, size_t * osamp)
{
  //(void)effp;   /* This parameter is not needed in this example */

  /* ensure that *osamp is a multiple of the number of channels. */
  *osamp -= *osamp % effp->out_signal.channels;

  /* Read up to *osamp samples into obuf; store the actual number read
   * back to *osamp */
  *osamp = in_ffsox_read(effp, obuf, *osamp);

  /* sox_read may return a number that is less than was requested; only if
   * 0 samples is returned does it indicate that end-of-file has been reached
   * or an error has occurred */
/*
  if (!*osamp && in->sox_errno)
    fprintf(stderr, "%s: %s\n", in->filename, in->sox_errstr);
*/

  return *osamp? SOX_SUCCESS : SOX_EOF;
}

/* A `stub' effect handler to handle inputting samples to the effects
 * chain; the only function needed for this example is `drain' */
static sox_effect_handler_t const * input_handler(void)
{
  static sox_effect_handler_t handler = {
    "input", NULL, SOX_EFF_MCHAN, NULL, NULL, NULL, input_drain, NULL, NULL, sizeof(track_t *)
  };
  return &handler;
}

static DWORD WINAPI in_ffsox_audio_thread(LPVOID p)
{
  track_t *track=p;

  sox_flow_effects(track->a.chain, NULL, NULL);
  track->a.done=1;
  samplebuf_signal_done(track->buf);

  return 0;
}

static DWORD WINAPI in_ffsox_video_thread(LPVOID p)
{
  track_t *track=p;
  video_t *v=&track->v;
  AVPacket pkt;

  while (NULL!=packet_queue_read(&v->qv, &pkt)) {
    in_ffsox_decode_video(track, &pkt);
    av_free_packet(&pkt);
  }

  return 0;
}

static DWORD WINAPI in_ffsox_dispatch_thread(LPVOID p)
{
  track_t *track=p;
  video_t *v=&track->v;
  audio_t *a=&track->a;
  packet_queue_t *qa=&v->qa;
  packet_queue_t *qv=&v->qv;
  AVPacket p1;

  HANDLE hThreadVideo=CreateThread(
    NULL,               // lpThreadAttributes (0 / CREATE_SUSPENDED)
    0,                  // dwStackSize
    in_ffsox_video_thread,    // lpStartAddress
    track,              // lpParameter
    0,                  // dwCreationFlags
    NULL                // lpThreadId
  );

  HANDLE hThreadAudio=CreateThread(
    NULL,               // lpThreadAttributes (0 / CREATE_SUSPENDED)
    0,                  // dwStackSize
    in_ffsox_audio_thread,    // lpStartAddress
    track,              // lpParameter
    0,                  // dwCreationFlags
    NULL                // lpThreadId
  );

#ifdef __DEBUG__
  fprintf(stderr, "start dispatching ...\n");
#endif
  for (;;) {
    for (;;) {
	  int of=packet_queue_overflow(qa)||packet_queue_overflow(qv);
	  int uf=packet_queue_underflow(qa)||packet_queue_underflow(qv);
	  int wait=of&&!uf;

	  wait=wait||v->pause;
	  wait=wait&&v->run;

	  if (wait)
        WaitForSingleObject(v->hEventWrite, INFINITE);
	  else
	    break;
	}

    if (! v->run)
      break;
    else if (av_read_frame(track->ic, &p1)<0)
      v->run=0;
    else if(p1.stream_index==a->s.id)
      packet_queue_write(qa, &p1);
    else if(p1.stream_index==v->s.id)
      packet_queue_write(qv, &p1);
    else
      av_free_packet(&p1);
  }

  packet_queue_signal_eof(qa);
  packet_queue_signal_eof(qv);
#ifdef __DEBUG__
  fprintf(stderr, "stop dispatching ...\n");
#endif
#ifdef __DEBUG__
  fprintf(stderr, "waiting for audio ...\n");
#endif
  WaitForSingleObject(hThreadAudio, INFINITE);
  CloseHandle(hThreadAudio);

#ifdef __DEBUG__
  fprintf(stderr, "waiting for video ...\n");
#endif
  WaitForSingleObject(hThreadVideo, INFINITE);
  CloseHandle(hThreadVideo);

#ifdef __DEBUG__
  fprintf(stderr, "done ...\n");
#endif
  return 0;
}

static void in_ffsox_close(track_t *track);

static track_t *in_ffsox_open(track_t *track, samplebuf_t *buf, const char *fn)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;
  AVFormatContext *ic;
  int id;

  memset(track, 0, sizeof *track);
  track->a.s.id=-1;
  track->v.s.id=-1;

#ifdef __DEBUG__
  fprintf(stderr, "*** Going to open AVFormat.\n");
#endif

  if (0!=av_open_input_file(&ic, fn, NULL, 0, NULL))
    return NULL;

  track->ic=ic;

  if (av_find_stream_info(ic)<0)
    goto error;

  for (id=0;id<ic->nb_streams;++id) {
    AVStream *stream=ic->streams[id];

    if (stream->codec->codec_type==CODEC_TYPE_AUDIO) {
      a->s.id=id;
      a->s.stream=stream;
      a->s.dec=stream->codec;
      a->s.codec=avcodec_find_decoder(a->s.dec->codec_id);
    }
    else if (stream->codec->codec_type==CODEC_TYPE_VIDEO
	    &&ini.winamp.vid) {

      v->s.id=id;
      v->s.stream=stream;
      v->s.dec=stream->codec;
      v->s.codec=avcodec_find_decoder(v->s.dec->codec_id);
    }
  }

  if (NULL==a->s.codec)
    goto error;

  samplebuf_open(buf, &ffsox_plugin, track, fn, a->s.dec->channels,
      a->s.dec->sample_rate,
	  av_get_bits_per_sample_format(a->s.dec->sample_fmt),
      ic->duration/1000);
  track->buf=buf;

  if (NULL!=ic->metadata)
    get_tags_md(ic->metadata, buf);

  if (NULL!=a->s.stream->metadata)
    get_tags_md(a->s.stream->metadata, buf);

  get_tags_fn(buf);
  in_ffsox_fill_format(track);

  return track;
error:
  in_ffsox_close(track);

  return NULL;
}

static track_t *in_ffsox_start(track_t *track, vout_t *vout, int (*gettime)(),
    void (*settime)(int), void (*showvis)())
{
  audio_t *a=&track->a;
  video_t *v=&track->v;
  DWORD WINAPI (*thread)(LPVOID p);

  a->start=0;
  a->done=0;
  a->settime=settime;
  samplebuf_start(track->buf);
  avcodec_open(a->s.dec, a->s.codec);
  a->s.time_base=1000.0*av_q2d(a->s.stream->time_base);

  if (NULL!=v->s.codec&&ini.winamp.vid) {
    AVStream *stream=v->s.stream;
    AVCodecContext *dec=v->s.dec;
    double aspectratio1, aspectratio2;

    avcodec_open(v->s.dec, v->s.codec);

    v->hEventWrite=CreateEvent(
      NULL,           // lpEventAttributes
      FALSE,          // bManualReset
      FALSE,          // bInitialState
      NULL            // lpName
    );

    v->run=1;
    v->pause=0;
    packet_queue_open(&v->qa, ini.queues.audio.min, ini.queues.audio.max,
	    v->hEventWrite);
    packet_queue_open(&v->qv, ini.queues.video.min, ini.queues.video.max,
	    v->hEventWrite);
    a->read_frame=in_ffsox_read_frame_v;
    thread=in_ffsox_dispatch_thread;

    ///////////////////////////////////////////////////////////////////////////
    v->width=v->s.dec->width;
    v->height=v->s.dec->height;
    v->in_pix_fmt=v->s.dec->pix_fmt;
    v->out_pix_fmt=PIX_FMT_YUV420P;

    aspectratio1
        =stream->sample_aspect_ratio.num?av_q2d(stream->sample_aspect_ratio)
        :dec->sample_aspect_ratio.num?av_q2d(dec->sample_aspect_ratio)
        :0.0
        ;
    aspectratio2=aspectratio1<=0.0?1.0:aspectratio1;

    vout_open(vout, v->width, v->height, 1.0/aspectratio2);
    vout_set_caption(vout, FFSOX);
    v->vout=vout;

    v->s.time_base=1000.0*av_q2d(stream->time_base);

	track->start_time=a->s.stream->start_time<v->s.stream->start_time
	    ?a->s.stream->start_time
	    :v->s.stream->start_time
		;
    v->step=1000.0/av_q2d(stream->r_frame_rate);
    v->gettime=gettime;
    timer_open(&v->timer, ini.plugin.timer_res);

    v->frame= avcodec_alloc_frame();

    if (v->out_pix_fmt!=v->in_pix_fmt) {
      v->img_convert_ctx=sws_getContext(
          v->width, v->height, v->in_pix_fmt,
          v->width, v->height, v->out_pix_fmt,
          SWS_BICUBIC, NULL, NULL, NULL);
    }

    avpicture_alloc(&v->out_pict, v->out_pix_fmt, v->width, v->height);
    avpicture_alloc(&v->in_pict, v->in_pix_fmt, v->width, v->height);

#ifdef __SYNC__
    // sync ////////////////////////////////
    InitializeCriticalSection(&v->sync.cs);

    v->sync.hEvent=CreateEvent(
      NULL,           // lpEventAttributes
      TRUE,           // bManualReset
      FALSE,          // bInitialState
      NULL            // lpName
    );

    v->sync.count=0;
    v->sync.a=0;
    v->sync.v=0;
#endif // __SYNC__
  }
  else {
	track->start_time=a->s.stream->start_time;
    a->read_frame=in_ffsox_read_frame_a;
    thread=in_ffsox_audio_thread;

    if (ini.winamp.vis)
      showvis();
  }

  a->chain=track_init_sox_chain(track, track->buf,
      av_get_bits_per_sample_format(a->s.dec->sample_fmt),
	  a->s.dec->sample_rate, input_handler);

  track->hThreadSoX=CreateThread(
    NULL,       // lpThreadAttributes (0 / CREATE_SUSPENDED)
    0,          // dwStackSize
    thread,     // lpStartAddress
    track,      // lpParameter
    0,          // dwCreationFlags
    NULL        // lpThreadId
  );

  return track;
}

static void in_ffsox_close(track_t *track)
{
  samplebuf_close(track->buf);

  if (NULL!=track->ic) {
    av_close_input_file(track->ic);
    track->ic=NULL;
#ifdef __DEBUG__
	fprintf(stderr, "*** AVFormat closed.\n");
#endif
  }

  memset(track, 0, sizeof *track);
}

static void in_ffsox_stop(track_t *track)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;

  if (NULL!=track->hThreadSoX) {
    WaitForSingleObject(track->hThreadSoX, INFINITE);
    CloseHandle(track->hThreadSoX);
  }

  if (NULL!=track->a.p1.data) {
    av_free_packet(&track->a.p1);
    memset(&track->a.p1, 0, sizeof track->a.p1);
  }

  if (NULL!=a->chain)
    sox_delete_effects_chain(a->chain);

  if (NULL!=v->s.codec) {
#ifdef __SYNC__
    if (NULL!=v->sync.hEvent)
      CloseHandle(v->sync.hEvent);

    DeleteCriticalSection(&v->sync.cs);
#endif // __SYNC__

    avpicture_free(&v->in_pict);
    avpicture_free(&v->out_pict);

    if (NULL!=v->img_convert_ctx)
      sws_freeContext(v->img_convert_ctx);

    if (NULL!=track->v.frame)
      av_free(track->v.frame);

    timer_close(&track->v.timer);

    if (NULL!=track->v.vout)
      vout_close(track->v.vout);

    packet_queue_close(&v->qv);
    packet_queue_close(&v->qa);

    if (NULL!=v->hEventWrite)
      CloseHandle(v->hEventWrite);

    if (NULL!=v->s.codec) {
      avcodec_close(v->s.dec);
      v->s.dec=NULL;
      v->s.codec=NULL;
    }
  }

  if (NULL!=a->s.codec) {
    avcodec_close(a->s.dec);
    a->s.dec=NULL;
    a->s.codec=NULL;
  }

  samplebuf_stop(track->buf);
}

static int filetype(const char *path, const char *ext)
{
  const char *rp=path+strlen(path);

  while (path<rp&&'.'!=rp[-1])
    --rp;

  return 0==strcmpi(rp,ext);
}

static track_t *in_ffsox_create_play(samplebuf_t *buf, const char *fn) {
#ifndef __FLAC_SOX__
  return ffmpeg_stub_valid?in_ffsox_open(&track_play, buf, fn):NULL;
#else // __FLAC_SOX__
  return filetype(fn, "flac")
    ?sox_plugin.create_play(buf, fn)
    :in_ffsox_open(&track_play, buf, fn)
    ;
#endif // __FLAC_SOX__
}

static track_t *in_ffsox_create_info(samplebuf_t *buf, const char *fn) {
#ifndef __FLAC_SOX__
  return ffmpeg_stub_valid?in_ffsox_open(&track_info, buf, fn):NULL;
#else // __FLAC_SOX__
  return filetype(fn, "flac")
    ?sox_plugin.create_info(buf, fn)
    :in_ffsox_open(&track_info, buf, fn)
    ;
#endif // __FLAC_SOX__
}

static void in_ffsox_kill(track_t *track)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;

  if (NULL!=v->s.codec) {
    v->run=0;
    packet_queue_signal_kill(&v->qa);
    packet_queue_signal_kill(&v->qv);
    timer_kill(&v->timer);
  }

  samplebuf_signal_done(track->buf);
}

static int in_ffsox_seek(track_t *track, int pos_ms)
{
  if (NULL==track->ic->iformat->read_seek)
    return 0;
  else {
    audio_t *a=&track->a;
    video_t *v=&track->v;
	stream_t *s=NULL!=v->s.codec?&v->s:&a->s;
    int64_t seek_target=av_rescale_q(1000*pos_ms, AV_TIME_BASE_Q,
        s->stream->time_base);

    in_ffsox_kill(track);

    if (NULL!=track->hThreadSoX) {
      WaitForSingleObject(track->hThreadSoX, INFINITE);
      CloseHandle(track->hThreadSoX);
      track->hThreadSoX=NULL;
    }

    return av_seek_frame(track->ic, s->id, seek_target, 0)<0?0:1;
#ifdef __XXX__
    return avformat_seek_file(track->ic, s->id, INT64_MIN, seek_target,
		INT64_MAX, 0)<0?0:1;
#endif // __XXX__
  }
}

static void in_ffsox_flush(track_t *track)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;

  samplebuf_reset(track->buf);
  a->done=0;
  a->start=0;

  if (NULL!=v->s.codec) {
    v->run=1;
    v->pause=0;
    packet_queue_reset(&v->qa);
    packet_queue_reset(&v->qv);
  }
}

static void in_ffsox_restart(track_t *track)
{
  audio_t *a=&track->a;
  video_t *v=&track->v;
  DWORD WINAPI (*thread)(LPVOID)=NULL!=v->s.codec
      ?in_ffsox_dispatch_thread
      :in_ffsox_audio_thread
	  ;

  track->hThreadSoX=CreateThread(
    NULL,       // lpThreadAttributes (0 / CREATE_SUSPENDED)
    0,          // dwStackSize
    thread,     // lpStartAddress
    track,      // lpParameter
    0,          // dwCreationFlags
    NULL        // lpThreadId
  );
}

static void in_ffsox_pause(track_t *track)
{
  video_t *v=&track->v;

  if (NULL!=v->s.codec) {
    v->qa.state=PACKET_QUEUE_PAUSE;
    v->qv.state=PACKET_QUEUE_PAUSE;
  }
}

static void in_ffsox_unpause(track_t *track)
{
  video_t *v=&track->v;

  if (NULL!=v->s.codec) {
    packet_queue_signal_work(&v->qa);
    packet_queue_signal_work(&v->qv);
  }
}

plugin_t ffsox_plugin = {
  .create_play=in_ffsox_create_play,
  .create_info=in_ffsox_create_info,
  .close=in_ffsox_close,
  .start=in_ffsox_start,
  .stop=in_ffsox_stop,
  .kill=in_ffsox_kill,
  .pause=in_ffsox_pause,
  .unpause=in_ffsox_unpause,
  .seek=in_ffsox_seek,
  .flush=in_ffsox_flush,
  .restart=in_ffsox_restart
};
