/*
 * prefs.c
 * Copyright (C) 2010 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "resources.h"
#include "plugin.h"
#include <commctrl.h>

//#define DEF_EXTS "AVI;MKV;MOV;MP3;MP4;RM;VOB;FLAC;FLV;WV;OGG"
#define DEF_EXTS "FLAC;MKV;OGG;WAV;WEBM;WV"
#define AC3_DECODER "ac3_decoder"
#define MP3_DECODER "mp3_decoder"

#define PREAMP_MIN (-12.0)
#define PREAMP_MAX (12.0)
#define PREAMP_SCALE (2.0)

#define PACKET_QUEUE_MIN 4
#define PACKET_QUEUE_MAX 8

#define PICTURE_QUEUE_MIN 4
#define PICTURE_QUEUE_MAX 8

#define LIBAVCODEC "libavcodec"
#define LIBA52 "liba52"
#define LIBMAD "libmad"

#define LOCATION "in_ffsox\\in_ffsox.ini"

ini_t ini;
char exts[4096];

typedef struct resample_prefs {
  const char *label;
  const int rate;
} resample_prefs_t;

typedef struct quality_prefs {
  const char *label;
  const char *quality;
} quality_prefs_t;

typedef struct limit_gain_prefs {
  const char *label;
  const char *gain;
} limit_gain_prefs_t;

static resample_prefs_t resample_prefs[] = {
  { "off", 0 },
  { "11025", 11025 },
  { "22050", 22050 },
  { "24000", 24000 },
  { "32000", 32000 },
  { "44100", 44100 },
  { "48000", 48000 },
  { "88200", 88200 },
  { "96000", 96000 },
  { "176400", 176400 },
  { "192000", 192000 },
  { NULL, 0 }
};

static quality_prefs_t quality_prefs[] = {
  { "default", "" },
  { "quick", "-q" },
  { "low", "-l" },
  { "medium", "-m" },
  { "high", "-h" },
  { "very high", "-v" },
  { NULL, NULL }
};

static limit_gain_prefs_t limit_gain_prefs[] = {
  { "0.01", "0.01" },
  { "0.02", "0.02" },
  { "0.03", "0.03" },
  { "0.04", "0.04" },
  { "0.05", "0.05" },
  { "0.06", "0.06" },
  { "0.07", "0.07" },
  { "0.08", "0.08" },
  { "0.09", "0.09" },
  { NULL, NULL }
};

static void get_exts()
{
  char *rp=ini.winamp.exts;
  char *mrp=ini.winamp.exts+strlen(ini.winamp.exts);
  char *wp=exts;
  char *mwp=exts+sizeof(exts)-4;

  memset(exts, 0, sizeof exts);

  while (rp<mrp&&wp<mwp) {
    if (';'==*rp) {
      *wp++=0;
      *wp++='X';
      *wp++=0;
      *wp=0;
      ++rp;
    }
    else
      *wp++=*rp++;
  }
}

static void fill_ini(const char *path, char *p) { 
    char buf[32];

    strcpy(p, LOCATION);

    // winamp
    GetPrivateProfileString("winamp", "exts", DEF_EXTS,
            ini.winamp.exts, sizeof ini.winamp.exts,
            ini.path);
    get_exts();
    ini.winamp.bps=GetPrivateProfileInt("winamp", "bps",
            16, ini.path);
    ini.winamp.vis=GetPrivateProfileInt("winamp", "vis",
            1, ini.path);
    ini.winamp.vid=GetPrivateProfileInt("winamp", "vid",
            1, ini.path);

    // rg
    ini.rg.mode=GetPrivateProfileInt("rg", "mode",
            1, ini.path);
    GetPrivateProfileString("rg", "preamp", "0",
            buf, sizeof buf, ini.path);
    //ini.rg.preamp=0.1*round(10.0*atof(buf));
    ini.rg.preamp=atof(buf);

    // sox
    GetPrivateProfileString("sox", "amplify", GAIN,
            ini.sox.amplify,
            sizeof ini.sox.amplify,
            ini.path);
    ini.sox.limit_switch=GetPrivateProfileInt("sox", "limit_switch",
            1, ini.path);
    GetPrivateProfileString("sox", "limit", "0.05",
            ini.sox.limit, sizeof ini.sox.limit,
            ini.path);
    ini.sox.sample_rate=GetPrivateProfileInt("sox", "sample_rate",
            0, ini.path);
#ifdef __ORIGINAL__
    ini.sox.very_high_quality=GetPrivateProfileInt("sox",
            "very_high_quality", 0, ini.path);
#else // ! __ORIGINAL__
    GetPrivateProfileString("sox", "quality", "", ini.sox.quality,
			sizeof ini.sox.quality, ini.path);
#endif // ! __ORIGINAL__
    ini.sox.largest_multiple=GetPrivateProfileInt("sox",
            "largest_multiple", 1, ini.path);
    ini.sox.dither=GetPrivateProfileInt("sox", "dither",
            1, ini.path);
    ini.sox.noise_shaping=GetPrivateProfileInt("sox", "noise_shaping",
            1, ini.path);

    // plugin
    ini.plugin.timer_res=GetPrivateProfileInt("plugin",
            "timer_res", 4, ini.path);
    ini.plugin.sox_decoder=GetPrivateProfileInt("plugin",
            "sox_decoder", 0, ini.path);
    ini.plugin.drop_frames.on=GetPrivateProfileInt("plugin",
            "drop_frames", 1, ini.path);
    ini.plugin.drop_frames.before=GetPrivateProfileInt("plugin",
            "drop_frames_before", -1000, ini.path);
    ini.plugin.drop_frames.after=GetPrivateProfileInt("plugin",
            "drop_frames_after", 1000, ini.path);
    GetPrivateProfileString("plugin", AC3_DECODER, LIBA52,
            ini.plugin._ac3_decoder,
            sizeof ini.plugin._ac3_decoder,
            ini.path);
    GetPrivateProfileString("plugin", MP3_DECODER, LIBMAD,
            ini.plugin._mp3_decoder,
            sizeof ini.plugin._mp3_decoder,
            ini.path);

    // queues
    ini.queues.audio.min=GetPrivateProfileInt("queues",
            "audio_min", PACKET_QUEUE_MIN, ini.path);
    ini.queues.audio.max=GetPrivateProfileInt("queues",
            "audio_max", PACKET_QUEUE_MAX, ini.path);
    ini.queues.video.min=GetPrivateProfileInt("queues",
            "video_min", PACKET_QUEUE_MIN, ini.path);
    ini.queues.video.max=GetPrivateProfileInt("queues",
            "video_max", PACKET_QUEUE_MAX, ini.path);
}

void load_ini()
{
  HMODULE module = GetModuleHandle (MODULE);

  if (!module||!GetModuleFileName(module, ini.path, sizeof ini.path))
    return;

  ini.p=ini.path+strlen(ini.path);

  while (ini.path<ini.p&&'\\'!=ini.p[-1])
    --ini.p;

  fill_ini(ini.path, ini.p);
}

static void save_ini()
{
    char buf[32];

    strcpy(ini.p, LOCATION);

    // winamp
    WritePrivateProfileString("winamp", "exts", ini.winamp.exts,
            ini.path);
    sprintf(buf, "%d", ini.winamp.bps);
    WritePrivateProfileString("winamp", "bps", buf, ini.path);
    sprintf(buf, "%d", ini.winamp.vis);
    WritePrivateProfileString("winamp", "vis", buf, ini.path);
    sprintf(buf, "%d", ini.winamp.vid);
    WritePrivateProfileString("winamp", "vid", buf, ini.path);

    // rg
    sprintf(buf, "%d", ini.rg.mode);
    WritePrivateProfileString("rg", "mode", buf, ini.path);
    sprintf(buf, "%.1f", 0.1*round(10.0*ini.rg.preamp));
    WritePrivateProfileString("rg", "preamp", buf, ini.path);

    // sox
    WritePrivateProfileString("sox", "amplify", ini.sox.amplify, ini.path);
    sprintf(buf, "%d", ini.sox.limit_switch);
    WritePrivateProfileString("sox", "limit_switch", buf, ini.path);
    WritePrivateProfileString("sox", "limit", ini.sox.limit, ini.path);
    sprintf(buf, "%d", ini.sox.sample_rate);
    WritePrivateProfileString("sox", "sample_rate", buf, ini.path);
#ifdef __ORIGINAL__
    sprintf(buf, "%d", ini.sox.very_high_quality);
    WritePrivateProfileString("sox", "very_high_quality", buf, ini.path);
#else // ! __ORIGINAL__
    WritePrivateProfileString("sox", "quality", ini.sox.quality, ini.path);
#endif // ! __ORIGINAL__
    sprintf(buf, "%d", ini.sox.largest_multiple);
    WritePrivateProfileString("sox", "largest_multiple", buf, ini.path);
    sprintf(buf, "%d", ini.sox.dither);
    WritePrivateProfileString("sox", "dither", buf, ini.path);
    sprintf(buf, "%d", ini.sox.noise_shaping);
    WritePrivateProfileString("sox", "noise_shaping", buf, ini.path);

    // plugin
    sprintf(buf, "%d", ini.plugin.timer_res);
    WritePrivateProfileString("plugin", "timer_res", buf, ini.path);
    sprintf(buf, "%d", ini.plugin.sox_decoder);
    WritePrivateProfileString("plugin", "sox_decoder", buf, ini.path);
    sprintf(buf, "%d", ini.plugin.drop_frames.on);
    WritePrivateProfileString("plugin", "drop_frames", buf, ini.path);
    sprintf(buf, "%d", ini.plugin.drop_frames.before);
    WritePrivateProfileString("plugin", "drop_frames_before", buf,
            ini.path);
    sprintf(buf, "%d", ini.plugin.drop_frames.after);
    WritePrivateProfileString("plugin", "drop_frames_after", buf,
            ini.path);
    WritePrivateProfileString("plugin", AC3_DECODER,
            ini.plugin._ac3_decoder, ini.path);
    WritePrivateProfileString("plugin", MP3_DECODER,
            ini.plugin._mp3_decoder, ini.path);

    // queues
    sprintf(buf, "%d", ini.queues.audio.min);
    WritePrivateProfileString("queues", "audio_min", buf, ini.path);
    sprintf(buf, "%d", ini.queues.audio.max);
    WritePrivateProfileString("queues", "audio_max", buf, ini.path);
    sprintf(buf, "%d", ini.queues.video.min);
    WritePrivateProfileString("queues", "video_min", buf, ini.path);
    sprintf(buf, "%d", ini.queues.video.max);
    WritePrivateProfileString("queues", "video_max", buf, ini.path);
}

static void enable_drop_frames(HWND hDlg, BOOL on)
{
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_PLUGIN_DROP_FRAMES_BEFORE), on);
    EnableWindow(GetDlgItem (hDlg, IDC_EDIT_PLUGIN_DROP_FRAMES_BEFORE), on);
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_PLUGIN_DROP_FRAMES_AFTER), on);
    EnableWindow(GetDlgItem (hDlg, IDC_EDIT_PLUGIN_DROP_FRAMES_AFTER), on);
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_PLUGIN_DROP_FRAMES_MS), on);
}

static void enable_dither(HWND hDlg, BOOL on)
{
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_SOX_NOISE_SHAPING), on);
    EnableWindow(GetDlgItem (hDlg, IDC_CHECKBOX_SOX_NOISE_SHAPING), on);
}

static void enable_resample(HWND hDlg, BOOL on)
{
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_SOX_LARGEST_MULTIPLE), on);
    EnableWindow(GetDlgItem (hDlg, IDC_CHECKBOX_SOX_LARGEST_MULTIPLE), on);
#ifdef __ORIGINAL__
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_SOX_VERY_HIGH_QUALITY), on);
    EnableWindow(GetDlgItem (hDlg, IDC_CHECKBOX_SOX_VERY_HIGH_QUALITY), on);
#else // ! __ORIGINAL__
    EnableWindow(GetDlgItem (hDlg, IDC_COMBO_SOX_RATE_QUALITY), on);
#endif // ! __ORIGINAL__
}

static void enable_vol(HWND hDlg, BOOL on)
{
#ifdef __ORIGINAL__
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_SOX_LIMIT), on);
    EnableWindow(GetDlgItem (hDlg, IDC_EDIT_SOX_LIMIT), on);
#else // ! __ORIGINAL__
    EnableWindow(GetDlgItem (hDlg, IDC_COMBO_SOX_LIMIT_GAIN),
	    on&&IsDlgButtonChecked (hDlg, LOWORD(IDC_CHECKBOX_SOX_LIMIT_SWITCH)));
#endif // ! __ORIGINAL__
}

static void enable_rg(HWND hDlg, BOOL on)
{
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_RG_PREAMP), on);
    EnableWindow(GetDlgItem (hDlg, IDC_STATIC_RG_PREAMP_DB), on);
    EnableWindow(GetDlgItem (hDlg, IDC_SLIDER_RG_PREAMP), on);
}

BOOL CALLBACK ConfigDlgProc (HWND hDlg, UINT message, WPARAM wParam,
        LPARAM lParam)
{
  char buf[32];
  int i, j;

  switch (message) {
    case WM_INITDIALOG:
      // winamp
      SetDlgItemText(hDlg, IDC_EDIT_WINAMP_EXTS, ini.winamp.exts);
      SendDlgItemMessage(hDlg, IDC_COMBO_WINAMP_BPS, CB_ADDSTRING, 0,
          (LPARAM)"16");
      SendDlgItemMessage(hDlg, IDC_COMBO_WINAMP_BPS, CB_ADDSTRING, 0,
          (LPARAM)"24");
      SendDlgItemMessage(hDlg, IDC_COMBO_WINAMP_BPS, CB_SETCURSEL,
          ini.winamp.bps/8-2, 0);
      CheckDlgButton(hDlg, IDC_CHECKBOX_WINAMP_VIS, ini.winamp.vis);
      CheckDlgButton(hDlg, IDC_CHECKBOX_WINAMP_VID, ini.winamp.vid);

      // rg
      SendDlgItemMessage(hDlg, IDC_COMBO_RG_MODE, CB_ADDSTRING, 0,
          (LPARAM)"off");
      SendDlgItemMessage(hDlg, IDC_COMBO_RG_MODE, CB_ADDSTRING, 0,
          (LPARAM)"track");
      SendDlgItemMessage(hDlg, IDC_COMBO_RG_MODE, CB_ADDSTRING, 0,
          (LPARAM)"album");
      SendDlgItemMessage(hDlg, IDC_COMBO_RG_MODE, CB_SETCURSEL,
          ini.rg.mode, 0);

      SendDlgItemMessage(hDlg, IDC_SLIDER_RG_PREAMP, TBM_SETRANGEMIN,
          TRUE, 0);
      SendDlgItemMessage(hDlg, IDC_SLIDER_RG_PREAMP, TBM_SETRANGEMAX,
          TRUE, PREAMP_SCALE*(PREAMP_MAX-PREAMP_MIN));
      SendDlgItemMessage(hDlg, IDC_SLIDER_RG_PREAMP, TBM_SETPOS,
          TRUE, PREAMP_SCALE*(ini.rg.preamp-PREAMP_MIN));

      // sox
#ifdef __ORIGINAL__
      SetDlgItemText(hDlg, IDC_EDIT_SOX_LIMIT, ini.sox.limit);
#else // ! __ORIGINAL__
      CheckDlgButton(hDlg, IDC_CHECKBOX_SOX_LIMIT_SWITCH,
          ini.sox.limit_switch);

      j=i=0;

	  while (NULL!=limit_gain_prefs[i].label) {
        SendDlgItemMessage(hDlg, IDC_COMBO_SOX_LIMIT_GAIN, CB_ADDSTRING, 0,
            (LPARAM)limit_gain_prefs[i].label);

        if (0==strcmp(limit_gain_prefs[i].gain,ini.sox.limit))
		  j=i;

		++i;
	  }

      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_LIMIT_GAIN, CB_SETCURSEL, j, 0);
#endif // ! __ORIGINAL__
      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_AMPLIFY, CB_ADDSTRING, 0,
          (LPARAM)GAIN);
      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_AMPLIFY, CB_ADDSTRING, 0,
          (LPARAM)VOL);
      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_AMPLIFY, CB_SETCURSEL,
          0==strcmp(GAIN,ini.sox.amplify)?0:1, 0);

      j=i=0;

	  while (NULL!=resample_prefs[i].label) {
        SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE, CB_ADDSTRING, 0,
            (LPARAM)resample_prefs[i].label);

        if (resample_prefs[i].rate==ini.sox.sample_rate)
		  j=i;

		++i;
	  }

      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE, CB_SETCURSEL, j, 0);

#ifdef __ORIGINAL__
      CheckDlgButton(hDlg, IDC_CHECKBOX_SOX_VERY_HIGH_QUALITY,
          ini.sox.very_high_quality);
#else // ! __ORIGINAL__
      j=i=0;

	  while (NULL!=quality_prefs[i].label) {
        SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE_QUALITY, CB_ADDSTRING, 0,
            (LPARAM)quality_prefs[i].label);

        if (0==strcmp(quality_prefs[i].quality,ini.sox.quality))
		  j=i;

		++i;
	  }

      SendDlgItemMessage(hDlg, IDC_COMBO_SOX_RATE_QUALITY, CB_SETCURSEL, j, 0);
#endif // ! __ORIGINAL__
      CheckDlgButton(hDlg, IDC_CHECKBOX_SOX_LARGEST_MULTIPLE,
          ini.sox.largest_multiple);
      CheckDlgButton(hDlg, IDC_CHECKBOX_SOX_DITHER, ini.sox.dither);
      CheckDlgButton(hDlg, IDC_CHECKBOX_SOX_NOISE_SHAPING,
          ini.sox.noise_shaping);

      // plugin
      sprintf(buf, "%d", ini.plugin.timer_res);
      SetDlgItemText(hDlg, IDC_EDIT_PLUGIN_TIMER_RES, buf);
      CheckDlgButton(hDlg, IDC_CHECKBOX_PLUGIN_SOX_DECODER,
              ini.plugin.sox_decoder);
      sprintf(buf, "%d", ini.plugin.drop_frames.before);
      CheckDlgButton(hDlg, IDC_CHECKBOX_PLUGIN_DROP_FRAMES,
              ini.plugin.drop_frames.on);
      sprintf(buf, "%d", ini.plugin.drop_frames.before);
      SetDlgItemText(hDlg, IDC_EDIT_PLUGIN_DROP_FRAMES_BEFORE, buf);
      sprintf(buf, "%d", ini.plugin.drop_frames.after);
      SetDlgItemText(hDlg, IDC_EDIT_PLUGIN_DROP_FRAMES_AFTER, buf);
      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_AC3_DECODER, CB_ADDSTRING, 0,
          (LPARAM)"libavcodec (float/16 bit)");
      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_AC3_DECODER, CB_ADDSTRING, 0,
          (LPARAM)"liba52 (double/32 bit)");
      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_AC3_DECODER, CB_SETCURSEL,
          0==strcmp(ini.plugin._ac3_decoder,LIBAVCODEC)?0:1, 0);
      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_MP3_DECODER, CB_ADDSTRING, 0,
          (LPARAM)"libavcodec (16 bit)");
      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_MP3_DECODER, CB_ADDSTRING, 0,
          (LPARAM)"libmad (29 bit)");
      SendDlgItemMessage(hDlg, IDC_COMBO_PLUGIN_MP3_DECODER, CB_SETCURSEL,
          0==strcmp(ini.plugin._mp3_decoder,LIBAVCODEC)?0:1, 0);

      // queues
      sprintf(buf, "%d", ini.queues.audio.min);
      SetDlgItemText(hDlg, IDC_EDIT_QUEUES_AUDIO_MIN, buf);
      sprintf(buf, "%d", ini.queues.audio.max);
      SetDlgItemText(hDlg, IDC_EDIT_QUEUES_AUDIO_MAX, buf);
      sprintf(buf, "%d", ini.queues.video.min);
      SetDlgItemText(hDlg, IDC_EDIT_QUEUES_VIDEO_MIN, buf);
      sprintf(buf, "%d", ini.queues.video.max);
      SetDlgItemText(hDlg, IDC_EDIT_QUEUES_VIDEO_MAX, buf);

      if (0==ini.sox.sample_rate)
        enable_resample(hDlg, FALSE);

      if (! ini.rg.mode)
        enable_rg(hDlg, FALSE);

      if (! ini.sox.dither)
        enable_dither(hDlg, FALSE);

      if (! ini.plugin.drop_frames.on)
        enable_drop_frames(hDlg, FALSE);

#ifdef __ORIGINAL__
      if (0==strcmp(ini.sox.amplify,GAIN))
        enable_vol(hDlg, FALSE);
#else // ! __ORIGINAL__
      enable_vol(hDlg, 0==strcmp(ini.sox.amplify,GAIN)?FALSE:TRUE);
#endif // ! __ORIGINAL__

      SetFocus(GetDlgItem (hDlg, IDC_EDIT_WINAMP_EXTS));
      break;
    case WM_NOTIFY:
      switch (LOWORD(wParam)) {
          case IDC_SLIDER_RG_PREAMP:
          sprintf(buf, "%.1f dB", 0.1*round(10.0*((double)SendDlgItemMessage(hDlg,
              IDC_SLIDER_RG_PREAMP, TBM_GETPOS, 0, 0)/PREAMP_SCALE+PREAMP_MIN)));
          SetDlgItemText(hDlg, IDC_STATIC_RG_PREAMP_DB, buf);
          break;
      }

      break;
    case WM_COMMAND:
      switch (LOWORD(wParam)) {
        case IDC_BUTTON_WINAMP_EXTS_DEF:
          SetDlgItemText(hDlg, IDC_EDIT_WINAMP_EXTS, DEF_EXTS);
          break;
        case IDC_COMBO_RG_MODE:
          enable_rg(hDlg, 0<SendDlgItemMessage(hDlg,
              IDC_COMBO_RG_MODE, CB_GETCURSEL, 0, 0)?TRUE:FALSE);
          break;
        case IDC_COMBO_SOX_RATE:
          enable_resample(hDlg, 0<SendDlgItemMessage(hDlg,
              IDC_COMBO_SOX_RATE, CB_GETCURSEL, 0, 0)?TRUE:FALSE);
          break;
        case IDC_CHECKBOX_SOX_DITHER:
          enable_dither(hDlg, IsDlgButtonChecked (hDlg,
              LOWORD(wParam))?TRUE:FALSE);
          break;
        case IDC_CHECKBOX_PLUGIN_DROP_FRAMES:
          enable_drop_frames(hDlg, IsDlgButtonChecked (hDlg,
              LOWORD(wParam))?TRUE:FALSE);
          break;
        case IDC_COMBO_SOX_AMPLIFY:
        case IDC_CHECKBOX_SOX_LIMIT_SWITCH:
          enable_vol(hDlg, 1==SendDlgItemMessage(hDlg,
              IDC_COMBO_SOX_AMPLIFY, CB_GETCURSEL, 0, 0)?TRUE:FALSE);
          break;
        case IDC_BUTTON_QUEUES_AUDIO_DEF:
          sprintf(buf, "%d", PACKET_QUEUE_MIN);
          SetDlgItemText(hDlg, IDC_EDIT_QUEUES_AUDIO_MIN, buf);
          sprintf(buf, "%d", PACKET_QUEUE_MAX);
          SetDlgItemText(hDlg, IDC_EDIT_QUEUES_AUDIO_MAX, buf);
          break;
        case IDC_BUTTON_QUEUES_VIDEO_DEF:
          sprintf(buf, "%d", PACKET_QUEUE_MIN);
          SetDlgItemText(hDlg, IDC_EDIT_QUEUES_VIDEO_MIN, buf);
          sprintf(buf, "%d", PACKET_QUEUE_MAX);
          SetDlgItemText(hDlg, IDC_EDIT_QUEUES_VIDEO_MAX, buf);
          break;
        case IDOK:
          // winamp
          GetDlgItemText(hDlg,IDC_EDIT_WINAMP_EXTS,ini.winamp.exts,
              sizeof ini.winamp.exts);
          ini.winamp.bps=8*(2+SendDlgItemMessage(hDlg,
              IDC_COMBO_WINAMP_BPS, CB_GETCURSEL, 0, 0));
          ini.winamp.vis=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_WINAMP_VIS)?1:0;
          ini.winamp.vid=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_WINAMP_VID)?1:0;

          // rg
          ini.rg.mode=SendDlgItemMessage(hDlg, IDC_COMBO_RG_MODE,
              CB_GETCURSEL, 0, 0);
          ini.rg.preamp=0.1*round(10.0*((double)SendDlgItemMessage(hDlg,
              IDC_SLIDER_RG_PREAMP, TBM_GETPOS, 0, 0)/PREAMP_SCALE+PREAMP_MIN));

          // sox
#ifdef __ORIGINAL__
          GetDlgItemText(hDlg,IDC_EDIT_SOX_LIMIT,ini.sox.limit,
              sizeof ini.sox.limit);
#else // ! __ORIGINAL__
          ini.sox.limit_switch=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_SOX_LIMIT_SWITCH)?1:0;
          strcpy(ini.sox.limit, limit_gain_prefs[SendDlgItemMessage(hDlg,
		      IDC_COMBO_SOX_LIMIT_GAIN, CB_GETCURSEL, 0, 0)].gain);
#endif // ! __ORIGINAL__
          ini.sox.sample_rate=resample_prefs[SendDlgItemMessage(hDlg,
		      IDC_COMBO_SOX_RATE, CB_GETCURSEL, 0, 0)].rate;
#ifdef __ORIGINAL__
          ini.sox.very_high_quality=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_SOX_VERY_HIGH_QUALITY)?1:0;
#else // ! __ORIGINAL__
          strcpy(ini.sox.quality, quality_prefs[SendDlgItemMessage(hDlg,
		      IDC_COMBO_SOX_RATE_QUALITY, CB_GETCURSEL, 0, 0)].quality);
#endif // ! __ORIGINAL__
          ini.sox.largest_multiple=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_SOX_LARGEST_MULTIPLE)?1:0;
          ini.sox.dither=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_SOX_DITHER)?1:0;
          ini.sox.noise_shaping=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_SOX_NOISE_SHAPING)?1:0;
          strcpy(ini.sox.amplify,0==SendDlgItemMessage(hDlg,
              IDC_COMBO_SOX_AMPLIFY,CB_GETCURSEL,0,0)?GAIN:VOL);

          // plugin
          GetDlgItemText(hDlg,IDC_EDIT_PLUGIN_TIMER_RES, buf, sizeof buf);
          ini.plugin.timer_res=atoi(buf);
          ini.plugin.sox_decoder=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_PLUGIN_SOX_DECODER)?1:0;
          ini.plugin.drop_frames.on=IsDlgButtonChecked (hDlg,
              IDC_CHECKBOX_PLUGIN_DROP_FRAMES)?1:0;
          GetDlgItemText(hDlg,IDC_EDIT_PLUGIN_DROP_FRAMES_BEFORE, buf,
              sizeof buf);
          ini.plugin.drop_frames.before=atoi(buf);
          GetDlgItemText(hDlg,IDC_EDIT_PLUGIN_DROP_FRAMES_AFTER, buf,
              sizeof buf);
          ini.plugin.drop_frames.after=atoi(buf);
          strcpy(ini.plugin._ac3_decoder,0==SendDlgItemMessage(hDlg,
              IDC_COMBO_PLUGIN_AC3_DECODER, CB_GETCURSEL, 0, 0)
              ?LIBAVCODEC:LIBA52);
          strcpy(ini.plugin._mp3_decoder,0==SendDlgItemMessage(hDlg,
              IDC_COMBO_PLUGIN_MP3_DECODER, CB_GETCURSEL, 0, 0)
              ?LIBAVCODEC:LIBMAD);

          // queues
          GetDlgItemText(hDlg,IDC_EDIT_QUEUES_AUDIO_MIN, buf, sizeof buf);
          ini.queues.audio.min=atoi(buf);
          GetDlgItemText(hDlg,IDC_EDIT_QUEUES_AUDIO_MAX, buf, sizeof buf);
          ini.queues.audio.max=atoi(buf);
          GetDlgItemText(hDlg,IDC_EDIT_QUEUES_VIDEO_MIN, buf, sizeof buf);
          ini.queues.video.min=atoi(buf);
          GetDlgItemText(hDlg,IDC_EDIT_QUEUES_VIDEO_MAX, buf, sizeof buf);
          ini.queues.video.max=atoi(buf);

          save_ini();
          get_exts();
          EndDialog (hDlg, 1);
          return TRUE;
        case IDCANCEL:
          EndDialog (hDlg, 0);
          return TRUE;
      }

      break;
  }

  return FALSE;
}
